/*
    By   Qianfeng Zhang ( frzhang@redhat.com )  November 2009 
*/

#include <qpid/client/FailoverManager.h>
#include <qpid/client/Session.h>
#include <qpid/client/Message.h>
#include <qpid/client/SubscriptionManager.h>

#include <iostream>
#include <sstream> 

#include "FailoverWorker.h"
#include "demo_common.h"

using namespace qpid;
using namespace qpid::client;
using namespace qpid::framing;

using namespace std;
using namespace demo;

namespace demo {

class AckListener :   public MessageListener, 
                      public demo::FailoverWorker
{
  public:
    AckListener(int station, string& hostIP, int nodes, int port);
    ~AckListener() {} ;
    void received(Message& message);    // called by MessageListner
    void execute(AsyncSession& session, bool isRetry); // called by FailoverManager
    void run();    //needed by the Listener_runner
    void sigHandle();   // this is used by the boost.signal

  private:
    int stationID; 
    Subscription subscription;

    int count; 
};


AckListener::AckListener(int station, string& hostIP, int nodes, int port) : FailoverWorker(station-1,hostIP,nodes,port)
{
     stationID = station; 

     count = 0; 
}

void AckListener::received(Message& message) 
{
     // cout << "ACK---" << message.getData() << endl;

     // manually acknowledge the message, but we do the acknowledgement every 100 messages to reduce
     // the lose in performance,  so it is possible that 100 messages could be received two times 
     // during broker switch-over 
     if  ( count++ % ACK_INTERVAL == 0 || ! system_running )  {
          subscription.acquire(subscription.getUnacquired());
          subscription.accept(subscription.getUnaccepted());
     }
}

void AckListener::execute(AsyncSession& session, bool isRetry)
{
    SubscriptionManager subs(session);

    // disable automatic acknowledgement, this is needed cause we don't want the messages be locked
    // during the broker switch-over
    subs.setAcquireMode(ACQUIRE_MODE_NOT_ACQUIRED);
    subs.setAcceptMode(ACCEPT_MODE_EXPLICIT);

    ostringstream os; 
    os << ORDER_ACK_QUEUE_PREFIX << stationID;   // setup the name of the queue to bind
    subscription = subs.subscribe(*this, os.str());
    subscription.setAutoAck(0);   // we will mannually ack the messages
    subs.run();
}

void AckListener::sigHandle()
{
     this->subscription.cancel(); 
}

void AckListener::run()
{
    FailoverManager fmanager(settings, this);

    cout << "Station " << stationID <<  "------ Order Acknowledgement Receiver Started ------- " << endl;

    try {
         fmanager.execute(*this);
         fmanager.close();
         std::cout << "Completed without error." << std::endl;
    } 
    catch (const std::exception& error) {
          std::cout << "Failure: " << error.what() << std::endl;
          system_running = 0;
    }

    cout << "Station " << stationID <<  "------ Order Acknowledgement Receiver Exited ------- " << endl;

    return;
}

class AckListener_runner
{
  private:
   AckListener *clsp;
  public:
   AckListener_runner(AckListener& lstner) : clsp(&lstner) {} ;
   ~AckListener_runner() {} ;
   void operator()() {
       clsp->run();
   };
};

class AckListener_sigHandler
{
  private:
   AckListener *clsp;
  public:
   AckListener_sigHandler(AckListener& lstner) : clsp(&lstner) {} ;
   ~AckListener_sigHandler() {} ;
   void operator()() {
       clsp->sigHandle();
   };
};


}  // namespace "demo"
