/*
    By   Qianfeng Zhang ( frzhang@redhat.com )  November 2009 
*/

#include <qpid/client/FailoverManager.h>
#include <qpid/client/Session.h>
#include <qpid/client/Message.h>
#include <qpid/client/SubscriptionManager.h>

#include <iostream>
#include <sstream> 

#include "FailoverWorker.h"
#include "demo_common.h"

using namespace qpid;
using namespace qpid::client;
using namespace qpid::framing;

using namespace std;
using namespace demo;

namespace demo {

class DataListener :   public MessageListener, 
                       public demo::FailoverWorker
{
  public:
    DataListener(int station, string& hostIP, int nodes, int port, void *priv);
    ~DataListener() {} ;
    void received(Message& message);    // called by MessageListner
    void execute(AsyncSession& session, bool isRetry); // called by FailoverManager
    void run();    //needed by the Listener_runner
    void sigHandle();   // this is used by the boost.signal

  private:
    int stationID; 
    Subscription subscription;

    int count; 
    struct stats_struct *statp;
};


DataListener::DataListener(int station, string& hostIP, int nodes, int port, void *priv) :
              FailoverWorker(station-1,hostIP,nodes,port)
{
     stationID = station; 

     count = 0; 
     statp = (struct stats_struct *) priv; 
}

void DataListener::received(Message& message) 
{
     /* do our processing of the message here */
     // cout << "DAT---" << message.getData() << endl;

     {
          boost::mutex::scoped_lock  l1(statp->lock);
          statp->orderClosed++; 
     }

     // manually acknowledge the message, but we do the acknowledgement every 100 messages to reduce
     // the lose in performance,  so it is possible that 100 messages could be received two times 
     // during broker switch-over 
     if  ( count++ % ACK_INTERVAL == 0 || ! system_running )  {     
          subscription.acquire(subscription.getUnacquired());   
          subscription.accept(subscription.getUnaccepted());   
     }

}

void DataListener::execute(AsyncSession& session, bool isRetry)
{
    SubscriptionManager subs(session);

    // disable automatic acknowledgement, this is needed cause we don't want the messages be locked
    // during the broker switch-over
    subs.setAcquireMode(ACQUIRE_MODE_NOT_ACQUIRED);
    subs.setAcceptMode(ACCEPT_MODE_EXPLICIT);

    ostringstream os; 
    os << ORDER_DAT_QUEUE_PREFIX << stationID;   // setup the name of the queue to bind
    subscription = subs.subscribe(*this, os.str());
    subscription.setAutoAck(0);

    subs.run();
}

void DataListener::sigHandle()
{
    this->subscription.cancel(); 
}

void DataListener::run()
{
    FailoverManager fmanager(settings, this);

    cout << "Station " << stationID << "------ Order Conclusion Receiver Started ------- " << endl;

    try {
        fmanager.execute(*this);
        fmanager.close();
        std::cout << "Completed without error." << std::endl;
    }  
    catch (const std::exception& error) {
          std::cout << "Failure: " << error.what() << std::endl;
          system_running = 0;
      }

    cout << "Station " << stationID << "------ Order Conclusion Receiver Exited ------- " << endl;

    return;
}

class DataListener_runner
{
  private:
   DataListener *clsp;
  public:
   DataListener_runner(DataListener& lstner) : clsp(&lstner) {} ;
   ~DataListener_runner() {} ;
   void operator()() {
       clsp->run();
   };
};

class DataListener_sigHandler
{
  private:
   DataListener *clsp;
  public:
   DataListener_sigHandler(DataListener& lstner) : clsp(&lstner) {} ;
   ~DataListener_sigHandler() {} ;
   void operator()() {
       clsp->sigHandle();
   };
};




}  // namespace "demo"
