/*
    By   Qianfeng Zhang ( frzhang@redhat.com )  November 2009 
*/


#include <qpid/client/FailoverManager.h>
#include <qpid/client/Session.h>
#include <qpid/client/Message.h>
#include <qpid/client/SubscriptionManager.h>
#include <qpid/framing/reply_exceptions.h>

#include <iostream>
#include <sstream> 

#include "FailoverWorker.h"
#include "demo_common.h"
#include "syncCommManagement.hpp"

#include <signal.h>
#include <unistd.h>
#include <boost/thread.hpp>
#include <boost/signal.hpp>



using namespace qpid;
using namespace qpid::client;
using namespace qpid::framing;

using namespace std;
using namespace demo;

namespace demo {

class Responser :   public MessageListener, 
                    public demo::FailoverWorker

{
  public:
    Responser(int stockGroup, string& hostIP, int nodes, int port);
    ~Responser() ;
    void received(Message& message);  // called by MessageListner
    void execute(AsyncSession& session, bool isRetry);  // called by FailoverManager
    void run();   // needed by the Responer_runner() 


    void setupCommManagement1(string& hostIP, int nodes, int port);   // for sending "Order confirmation" messaging
    void setupCommManagement2(string& hostIP, int nodes, int port);   // for sending "Order Conclusion" messaging 

    void sendOrderConfirmation(Message& message, char * dstExchg);
    void sendOrderConclusion(Message& message, char * dstExchg); 

    void sigHandle();   // this is used by the boost.signal
  private:
    int stockGID; 
    Subscription subscription;

    syncCommManagement  *CM1p;   // for communication management of "Order confirmation" messaging
    syncCommManagement  *CM2p;   // for communication management of "Order Conclusion" messaging 

    int count; 

};


Responser::Responser(int stockGroup, string& hostIP, int nodes, int port) : FailoverWorker(stockGroup-1,hostIP,nodes,port)
{
     stockGID = stockGroup; 

     CM1p=NULL;
     CM2p=NULL; 

     count = 0; 
}

Responser::~Responser()
{
    if ( CM1p) 
         delete CM1p;
    if ( CM2p)
         delete CM2p; 
}


void Responser::received(Message & message) 
{
     Message ack_response,data_response; 
     ostringstream  ack_routingKey, data_routingKey;
     string info;
     uint stationID;
     uint stockID; 

     stationID = message.getHeaders().getAsInt("station");
     stockID = message.getHeaders().getAsInt("stock");
     
     info = message.getData(); 
     
     cout << "REQ---" << info << endl; 

     /* send a acknowledge message so that the sender know the request is received */
     ack_response.setData(info); 
     ack_response.getHeaders().setInt("stock", stockID);
     ack_routingKey << "Key_" << stationID; 
     ack_response.getDeliveryProperties().setRoutingKey(ack_routingKey.str());
     this->sendOrderConfirmation(ack_response, ORDER_ACK_EXCHG_NAME);
     

     /* do something here */
     // usleep(2); 

     /* send a Order conclusion message so that the sender know the result of their requests */
     data_response.setData(info); 
     data_response.getHeaders().setInt("stock", stockID);
     data_routingKey << "Key_" << stationID ;  
     data_response.getDeliveryProperties().setRoutingKey(data_routingKey.str());
     this->sendOrderConclusion(data_response, ORDER_DAT_EXCHG_NAME);

     // manually acknowledge the message, but we do the acknowledgement every 100 messages to reduce
     // the lose in performance,  so it is possible that 100 messages could be received two times 
     // during broker switch-over 
     if  ( count++ % ACK_INTERVAL == 0 || ! system_running )  {
          subscription.acquire(subscription.getUnacquired());
          subscription.accept(subscription.getUnaccepted());
     }

}

void Responser::execute(AsyncSession& session, bool isRetry)
{
    SubscriptionManager subs(session);

    // disable automatic acknowledgement, this is needed cause we don't want the messages be locked
    // during the broker switch-over
    subs.setAcquireMode(ACQUIRE_MODE_NOT_ACQUIRED);
    subs.setAcceptMode(ACCEPT_MODE_EXPLICIT);

    ostringstream  os; 
    os << ORDER_REQ_QUEUE_PREFIX << stockGID;   // setup the name of the queue to bind
    subscription = subs.subscribe(*this, os.str());
    subscription.setAutoAck(0);
    subs.run();
}

void Responser::setupCommManagement1(string& hostIP, int nodes, int port)
{
     CM1p =  new syncCommManagement(stockGID-1, hostIP, nodes, port); 
     CM1p->initConnect();
}

void Responser::setupCommManagement2(string& hostIP, int nodes, int port)
{
     CM2p =  new syncCommManagement(stockGID-1, hostIP, nodes, port); 
     CM2p->initConnect();
}


// this function will automatically switch over brokers and re-send the message when transport fails 
void Responser::sendOrderConfirmation(Message& message, char * dstExchg)
{
     bool completed = false; 
     bool reconnect_needed = false; 

     while ( system_running && !completed) {
         try {
              if  ( ! reconnect_needed )  {
                    CM1p->curSession().messageTransfer(arg::content=message,  arg::destination=dstExchg);
              }
              else  {
                  CM1p->reconnect();
                  reconnect_needed = false; 
                  continue;  
              }
              completed = true;  
         }
         catch (const TransportFailure&) {
              if ( ! reconnect_needed ) {
                     reconnect_needed = true;    
              }
         }
         catch (const ResourceLimitExceededException& ) {
              cerr << "ResourceLimitExceeded while sending order confirmation" << endl;
              usleep(SEND_RETRY_INTERVAL);    // re-try after xxx micro-seconds 
              CM1p->resetSession();
         }
     }     
}

// this function will automatically switch over brokers and re-send the message when transport fails 
void Responser::sendOrderConclusion(Message& message, char * dstExchg)
{
     bool completed = false;
     bool reconnect_needed = false;

     while ( system_running && !completed) {
         try {
              if  ( ! reconnect_needed )  {
                    CM2p->curSession().messageTransfer(arg::content=message,  arg::destination=dstExchg);
              }
              else  {
                  CM2p->reconnect();
                  reconnect_needed = false;
                  continue;
              }
              completed = true;
         }
         catch(const TransportFailure&) {
              if ( ! reconnect_needed ) {
                     reconnect_needed = true;
              }
         }
         catch (const ResourceLimitExceededException& ) {
              cerr << "ResourceLimitExceeded while sending order conclusion" << endl;
              usleep(SEND_RETRY_INTERVAL);    // re-try after xxx micro-seconds 
              CM2p->resetSession();
         }
     } 
}

// handler for  boost.signal
void Responser::sigHandle()
{
   this->subscription.cancel(); 
}

void Responser::run()
{
    FailoverManager fmanager(settings, this);

    cout << "Stock Group " << stockGID << "------ Order Handler Started ------- " << endl;

    try {
        fmanager.execute(*this);
        fmanager.close();
        std::cout << "Completed without error." << std::endl;
    } 
    catch (const std::exception& error) {
          std::cout << "Failure: " << error.what() << std::endl;
          system_running = 0; 
      }

    cout << "Stock Group " << stockGID << "------ Order Handler Exited ------- " << endl;

   return;
}

class  Responser_runner
{
  private:
   Responser *clsp;
  public:
   Responser_runner(Responser& resp) : clsp(&resp) {} ;
   ~Responser_runner() {};
   void operator()() {
       clsp->run();
   };
};


class  Responser_sigHandler
{
  private:
   Responser *clsp;
  public: 
    Responser_sigHandler(Responser& resp) : clsp(&resp) {} ; 
    ~Responser_sigHandler() {}; 
    void operator()() {
        clsp->sigHandle(); 
    };
};


}  // namespace "demo"


int system_running;

static void my_quit(int sig);

static boost::thread *thrd1p;

static boost::signal<void ()> mysig;

int main(int argc,  char *argv[])
{
   int stockGroup;
   string brokerIP;
   int nodes;
   int port;    

   if ( argc != 5 ) {
       cout << "Usage: " << argv[0] << " <stockGroup> <brokerIP> <nodes> <port> " << endl;
       exit(2);
   } 

   stockGroup = atoi(argv[1]);
   if (stockGroup > MAX_STOCK_GROUP || stockGroup <= 0 )  
       stockGroup = stockGroup % MAX_STOCK_GROUP + 1; 
   brokerIP = argv[2];
   nodes = atoi(argv[3]);
   port = atoi(argv[4]);


   demo::Responser myresponser(stockGroup, brokerIP, nodes, port);

   myresponser.setupCommManagement1(brokerIP, nodes, port);     // could be different clusters of brokers
   myresponser.setupCommManagement2(brokerIP, nodes, port);

   mysig.connect(Responser_sigHandler(myresponser)); 

   system_running = 1;

   thrd1p = new boost::thread(Responser_runner(myresponser));
   
   signal(SIGINT, my_quit);
   signal(SIGQUIT, SIG_IGN);
   signal(SIGTERM, SIG_IGN);

   while ( system_running )
         sleep(1);

   thrd1p->join();

   delete thrd1p; 
}

static void my_quit(int sig)
{
    // cout << "Signal " << sig << " occurred " << endl;
    system_running = 0 ;
    mysig();
}


