/*
    By   Qianfeng Zhang ( frzhang@redhat.com )  November 2009 
*/


#include <qpid/client/FailoverManager.h>
#include <qpid/client/Session.h>
#include <qpid/client/AsyncSession.h>
#include <qpid/client/Message.h>
#include <qpid/client/MessageReplayTracker.h>
#include <qpid/Exception.h>
#include <qpid/framing/reply_exceptions.h>

#include <iostream>
#include <sstream>

#include <unistd.h>
#include "demo_common.h"
#include "asynCommManagement.hpp"

using namespace qpid;
using namespace qpid::client;
using namespace qpid::framing;

using namespace demo;
using namespace std;

namespace demo {

class Sender 
{
  public:
    Sender(int station, string& hostIP, int nodes, int port);
    ~Sender() {};
    void execute();   // called by FailoverManager
    uint getSent();
    void run();    // needed by Sender_runner
    void sigHandle();   // this is used by the boost.signal

    void setupCommManagement(string& hostIP, int nodes, int port);   // for sending "Order request" messages
    void sendOrderRequest(Message& message, char * dstExchg);

  private:
    int stationID;
    uint sent;

    asynCommManagement  *CMp;   // for communication management of "Order request" messaging
    
    int  getStockID();   
};

Sender::Sender(int station, string& hostIP, int nodes, int port) 
{
    stationID = station; 
    sent = 0; 
}

void Sender::execute()
{
    Message message;

    while ( system_running ) {
        ostringstream message_data;
        ostringstream  routingKey;
        int stockID, stockGID; 

        stockID = getStockID(); 
        stockGID = stockID % MAX_STOCK_GROUP + 1; 
        message_data << "STOCK TRADE REQUEST: NO." << sent << " Stock." << stockID << " Station." << stationID; 
        message.setData(message_data.str());
        message.getHeaders().setInt("stock", stockID); 
        message.getHeaders().setInt("station", stationID);
        routingKey << "Key_" << stockGID; 
        message.getDeliveryProperties().setRoutingKey(routingKey.str());

        this->sendOrderRequest(message, ORDER_REQ_EXCHG_NAME);
        sent++;   // go forward to a new message 
    }
}

uint Sender::getSent()
{
    return sent;
}

void Sender::setupCommManagement(string& hostIP, int nodes, int port)
{
     CMp =  new asynCommManagement(stationID-1, hostIP, nodes, port);
     CMp->initConnect();
}


// this function will automatically switch over brokers and re-send the message when transport fails 
void Sender::sendOrderRequest(Message& message, char * dstExchg)
{
     bool completed = false; 
     bool reconnect_needed = false; 
     bool just_reconnected = false; 
     bool resetSession_needed = false; 

     while ( system_running && !completed) {
         try {
              if  ( resetSession_needed ) {
                  CMp->resetSession();            // this is necessary cause the session is detached by the server side
                  CMp->getTracker().replay(CMp->curSession());
                  resetSession_needed = false; 
                  completed = true; 
                  break;       // all messages are sent by ->replay() 
              }
              if  ( ! reconnect_needed )  {
                    if ( just_reconnected )  { //using the new session, and replay the old messages
                        CMp->getTracker().replay(CMp->curSession());
                        just_reconnected = false; 
                        completed = true;
                        break;   // all messages are sent by ->replay()
                    }
                    else
                       CMp->getTracker().send(message, dstExchg);
              }
              else  {
                  CMp->reconnect();
                  just_reconnected = true;
                  reconnect_needed = false;
                  continue;  
              }
              completed = true;
         }
         catch (const TransportFailure&) {
              cout << "TransportFailure while sending order request" << endl;
              reconnect_needed = true;
              resetSession_needed = false; 
         }
         catch (const ResourceLimitExceededException& ) {
              cerr << "ResourceLimitExceeded while sending order request" << endl;
              usleep(SEND_RETRY_INTERVAL);    // re-try after xxx micro-seconds 
              resetSession_needed = true; 
         }
     }
}  


void Sender::run()
{
   cout << "Station " << stationID <<  "------ Station Sender Started ------- " << endl; 

   try {
        this->execute();
   }
   catch (const std::exception& error) {
          std::cout << "Failure: " << error.what() << std::endl;
          system_running = 0; 
   }

   cout << "Station " << stationID <<  "------ Station Sender Exited ------- " << endl; 
}

void Sender::sigHandle()
{
  
}

int Sender::getStockID()
{
    return( rand() % MAXSTOCKNUM ) ; 
}

class Sender_runner
{
  private:
   class Sender *clsp;
  public:
   Sender_runner(Sender& snder) : clsp(&snder) {} ;
   ~Sender_runner()  {};
   void operator()() {
       clsp->run();
   };
};

class  Sender_sigHandler
{
  private:
    Sender *clsp;
  public:
    Sender_sigHandler(Sender& snder) : clsp(&snder) {} ;
   ~Sender_sigHandler() {} ;
   void operator()() {
       clsp->sigHandle();
   };
};



}   // end of namespace "demo"
