/*
    By   Qianfeng Zhang ( frzhang@redhat.com )  November 2009 

     This is a multi-threaded sender, with each thread stands for a  "sub-station" 
     Each thread:   1)  Generate "request" message with random "stock id", the "stock id" is mapped
                        to one "stock group id",  the request message will be received by one queue
                        bound to the "stock group id". Each "request" message is also tagged with one
                        "sub-station ID" so that a "response" message corresponding to this request
                        might be sent.  
                    2)  Receive "response" message from a queue bound to the "sub-station ID"
                    3)  handle the broker fail-over
 */

#include <iostream>
#include <sstream>

#include "demo_sender.hpp"
#include "demo_ack_receiver.hpp"
#include "demo_data_receiver.hpp"
#include "demo_data_stat.hpp"

#include <signal.h>
#include <boost/thread.hpp>
#include <boost/signal.hpp>

using namespace demo;
using namespace std;


int system_running;

static struct stats_struct  myStat;

static void my_quit(int sig);
static void set_process_fifo();

static boost::thread *thrd1p, *thrd2p, *thrd3p, *thrd4p;

static boost::signal<void ()> mysig;

int main(int argc,  char *argv[])
{
   int station;
   string brokerIP;
   int nodes;
   int port;    

   if ( argc != 5 ) {
       cout << "Usage: " << argv[0] << " <stationNumber> <brokerIP> <nodes> <port> " << endl;
       exit(2);
   } 
   
   station = atoi(argv[1]);
   if ( station <=0 || station > MAX_STATION ) 
      station = station % MAX_STATION + 1; 
   brokerIP = argv[2];
   nodes = atoi(argv[3]);
   port = atoi(argv[4]);

   demo::Sender  myReqSender(station, brokerIP, nodes, port);

   myReqSender.setupCommManagement(brokerIP, nodes, port);

   demo::AckListener myAckReceiver(station, brokerIP, nodes, port);
   demo::DataListener myDataReceiver(station, brokerIP, nodes, port, &myStat);
   demo::DataStats  myStatsWorker(&myStat); 

   mysig.connect(AckListener_sigHandler(myAckReceiver));
   mysig.connect(DataListener_sigHandler(myDataReceiver));
   mysig.connect(DataStats_sigHandler(myStatsWorker)); 
   mysig.connect(Sender_sigHandler(myReqSender)); 

   system_running = 1; 

   thrd1p = new boost::thread(demo::Sender_runner(myReqSender));
   thrd2p = new boost::thread(demo::DataStats_runner(myStatsWorker));
   thrd3p = new boost::thread(demo::AckListener_runner(myAckReceiver));
   thrd4p = new boost::thread(demo::DataListener_runner(myDataReceiver));

   signal(SIGINT, my_quit);
   signal(SIGQUIT, SIG_IGN);
   signal(SIGTERM, SIG_IGN);

   while ( system_running )  {
         sleep(1); 
   }
     
   thrd1p->join();
   thrd2p->join(); 
   thrd3p->join(); 
   thrd4p->join(); 

   delete thrd1p;
   delete thrd2p;
   delete thrd3p;
   delete thrd4p;
}


static void my_quit(int sig)
{
    // cout << "Signal " << sig << " caught " << endl;
    system_running = 0 ; 
    mysig(); 
}


